/*
 * =================================================================================
 * Copyright (C) 2013 Martin Albedinsky [Wolf-ITechnologies]
 * =================================================================================
 * Licensed under the Apache License, Version 2.0 or later (further "License" only);
 * ---------------------------------------------------------------------------------
 * You may use this file only in compliance with the License. More details and copy
 * of this License you may obtain at
 * 
 * 		http://www.apache.org/licenses/LICENSE-2.0
 * 
 * You can redistribute, modify or publish any part of the code written in this
 * file but as it is described in the License, the software distributed under the 
 * License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES or CONDITIONS OF
 * ANY KIND.
 * 
 * See the License for the specific language governing permissions and limitations
 * under the License.
 * =================================================================================
 */
package com.wit.and.database;

import android.content.ContentProvider;
import android.content.ContentValues;
import android.content.Context;
import android.content.UriMatcher;
import android.database.Cursor;
import android.database.sqlite.SQLiteDatabase;
import android.database.sqlite.SQLiteOpenHelper;
import android.net.Uri;
import android.util.Log;
import android.util.SparseArray;

import com.wit.and.database.annotations.DatabaseAuthority;
import com.wit.and.database.annotations.DatabaseEntities;
import com.wit.and.database.annotations.DatabaseName;
import com.wit.and.database.annotations.DatabaseVersion;

import java.io.File;

/**
 * <h4>Class Overview</h4>
 * <p>
 * </p>
 *
 * @author Martin Albedinsky
 */
public abstract class BaseDatabaseProvider extends ContentProvider {

    /**
     * Constants =============================
     */

    /**
     * Log TAG.
     */
    private static final String TAG = BaseDatabaseProvider.class.getSimpleName();

    /**
     * Indicates if debug private output trough log-cat is enabled.
     */
    // private static final boolean DEBUG = true;

    /**
     * Indicates if logging for user output trough log-cat is enabled.
     */
    private static final boolean USER_LOG = true;

	/**
	 *
	 */
	private static final UriMatcher URI_MATCHER = new UriMatcher(UriMatcher.NO_MATCH);

	/**
	 *
	 */
	public static final int MAX_CACHED_ENTITIES = 30;

    /**
     * Enums =================================
     */

    /**
     * Static members ========================
     */

	/**
	 *
	 */
    private static int sMaxCachedEntitiesCount = 10;

    /**
     * Members ===============================
     */

    private String mDatabaseName;

    private String mDatabaseAuthority;

    private int mDatabaseVersion;

    private SQLiteOpenHelper mSQLiteHelper;

    /**
     * Listeners -----------------------------
     */

    /**
     * Arrays --------------------------------
     */


    private static final SparseArray<Class<? extends BaseDatabaseEntity>> ENTITIES = new SparseArray<Class<? extends BaseDatabaseEntity>>();

    private static final SparseArray<BaseDatabaseEntity> CACHED_ENTITIES = new SparseArray<BaseDatabaseEntity>();

    /**
     * Booleans ------------------------------
     */

    /**
     * Constructors ==========================
     */

    /**
     * <p>
     * </p>
     */
    protected BaseDatabaseProvider() {
        super();
	    /**
	     * Process class annotations.
	     */
	    final Class<? extends BaseDatabaseProvider> providerClass = getClass();

	    // Retrieve database name.
	    if (providerClass.isAnnotationPresent(DatabaseName.class)) {
		    this.mDatabaseName = providerClass.getAnnotation(DatabaseName.class).value();
	    }

	    // Retrieve database authority.
	    if (providerClass.isAnnotationPresent(DatabaseAuthority.class)) {
		    this.mDatabaseAuthority = providerClass.getAnnotation(DatabaseAuthority.class).value();
	    }

	    // Retrieve database version.
	    if (providerClass.isAnnotationPresent(DatabaseVersion.class)) {
		    this.mDatabaseVersion = providerClass.getAnnotation(DatabaseVersion.class).value();
	    }
    }

	/**
	 *
	 * @param databaseAuthority
	 * @param databaseName
	 * @param databaseVersion
	 */
	protected BaseDatabaseProvider(String databaseAuthority, String databaseName, int databaseVersion) {
		super();
		this.mDatabaseName = databaseName;
		this.mDatabaseAuthority = databaseAuthority;
		this.mDatabaseVersion = databaseVersion;
	}

    /**
     * Methods ===============================
     */

    /**
     * Public --------------------------------
     */

    /**
     */
    @Override
    public boolean onCreate() {
        // Dispatch to prepare tables.
        this.dispatchPrepareTables();

        // Initialize SQLite open helper.
        this.mSQLiteHelper = new SQLiteOpenHelper(getContext(), mDatabaseName, null, mDatabaseVersion) {

            /**
             */
            @Override
            public void onCreate(SQLiteDatabase db) {
                if (USER_LOG)
                    Log.i(TAG, "Creating new database('" + mDatabaseName + "') in version(" + mDatabaseVersion + ").");

                // Perform create process of database.
                performCreateDatabase(db);

                if (USER_LOG)
                    Log.i(TAG, "New database('" + mDatabaseName + "') was successfully created.");
                dispatchDatabaseCreated();
            }

            /**
             */
            @Override
            public void onUpgrade(SQLiteDatabase db, int oldVersion, int newVersion) {
                performUpgradeDatabase(db, oldVersion, newVersion);
            }

            /**
             */
            @Override
            public void onOpen(SQLiteDatabase db) {
                super.onOpen(db);
                dispatchDatabaseOpen(db);
            }
        };

        // Invoke to crate database.
        mSQLiteHelper.getWritableDatabase();

        if (USER_LOG) {
            Log.i(TAG, "Database provider was successfully created.");
        }
        return true;
    }

    /**
     */
    @Override
    public Cursor query(Uri uri, String[] projection, String selection, String[] selectionArgs, String sortOrder) {
        BaseDatabaseEntity entity = getEntityByUri(uri);
        if (entity != null) {
            cacheEntity(entity);
            return entity.dispatchQuery(getSQLiteHelper().getReadableDatabase(), projection, selection, selectionArgs, sortOrder);
        }
        return null;
    }

    /**
     */
    @Override
    public Uri insert(Uri uri, ContentValues values) {
        BaseDatabaseEntity entity = getEntityByUri(uri);
        if (entity != null) {
            cacheEntity(entity);
            return entity.dispatchInsert(getSQLiteHelper().getWritableDatabase(), values);
        }
        return null;
    }

    /**
     */
    @Override
    public int bulkInsert(Uri uri, ContentValues[] values) {
        int rows = -1;
        BaseDatabaseEntity entity = getEntityByUri(uri);
        if (entity != null) {
            cacheEntity(entity);
            rows = entity.dispatchBulkInsert(getSQLiteHelper().getWritableDatabase(), values);
        }
        return (rows != -1) ? rows : super.bulkInsert(uri, values);
    }

    /**
     */
    @Override
    public int delete(Uri uri, String selection, String[] selectionArgs) {
        BaseDatabaseEntity entity = getEntityByUri(uri);
        if (entity != null) {
            cacheEntity(entity);
            return entity.dispatchDelete(getSQLiteHelper().getWritableDatabase(), selection, selectionArgs);
        }
        return -1;
    }

    /**
     */
    @Override
    public int update(Uri uri, ContentValues values, String selection, String[] selectionArgs) {
        BaseDatabaseEntity entity = getEntityByUri(uri);
        if (entity != null) {
            cacheEntity(entity);
            return entity.dispatchUpdate(getSQLiteHelper().getWritableDatabase(), values, selection, selectionArgs);
        }
        return -1;
    }

    /**
     * Getters + Setters ---------------------
     */

	/**
	 * <p>
	 * </p>
	 *
	 * @param count
	 */
	public static void setMaxCachedEntitiesCount(int count) {
		if (count >= 0) {
			sMaxCachedEntitiesCount = (count <= MAX_CACHED_ENTITIES) ? count : MAX_CACHED_ENTITIES;
		}
	}

	/**
	 * <p>
	 * </p>
	 *
	 * @return
	 */
	public static int getMaxCachedEntitiesCount() {
		return sMaxCachedEntitiesCount;
	}

    /**
     * <p>
     * </p>
     *
     * @return
     */
    public String getDatabaseName() {
        return mDatabaseName;
    }

    /**
     * <p>
     * </p>
     *
     * @return
     */
    public int getDatabaseVersion() {
        return mDatabaseVersion;
    }

    /**
     * <p>
     * </p>
     *
     * @return
     */
    public String getDatabaseAuthority() {
        return mDatabaseAuthority;
    }

    /**
     */
    @Override
    public String getType(Uri uri) {
        return "vnd.android.cursor.dir/" + mDatabaseAuthority + "/" + "table";
    }

    /**
     * Protected -----------------------------
     */

	/**
	 *
	 * @param classOfEntity
	 */
	protected void assignEntity(Class<? extends BaseDatabaseEntity> classOfEntity) {
		if (classOfEntity != null) {
			BaseDatabaseEntity entity = null;
			try {
				entity = classOfEntity.newInstance();
			} catch (InstantiationException e) {
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			}

			if (entity != null) {
				// Check for already assigned entity.
				if (URI_MATCHER.match(entity.getContentUri()) != -1) {
					if (USER_LOG)
						Log.v(TAG, "Entity with URI(" + entity.getContentUri() + ") is already assigned.");
				} else {
					final int entityCode = ENTITIES.size();
					URI_MATCHER.addURI(mDatabaseAuthority, entity.getName(), entityCode);
					ENTITIES.append(entityCode, classOfEntity);
				}
			}
		}
	}

	/**
	 * <p>
	 * </p>
	 *
	 * @return
	 */
	protected SparseArray<Class<? extends BaseDatabaseEntity>> getAssignedEntities() {
		return ENTITIES;
	}

	/**
	 * <p>
	 * </p>
	 *
	 * @return
	 */
	protected UriMatcher getUriMatcher() {
		return URI_MATCHER;
	}

    /**
     * <p>
     * </p>
     *
     * @return
     */
    protected SQLiteOpenHelper getSQLiteHelper() {
        return mSQLiteHelper;
    }

    /**
     * <p>
     * </p>
     *
     * @param db
     */
    protected void onCreateDatabase(SQLiteDatabase db) {
    }

	/**
	 * <p>
	 * </p>
	 */
	protected void onPrepareEntities() {
	}

    /**
     * <p>
     * </p>
     *
     * @param db
     * @param oldVersion
     * @param newVersion
     */
    protected void onUpgradeDatabase(SQLiteDatabase db, int oldVersion, int newVersion) {
        if (oldVersion != newVersion) {
            if (USER_LOG)
                Log.i(TAG, "Request to upgrade database from(" + oldVersion + ") to(" + newVersion + "). Dropping old database and creating new one.");

            // Get database file and delete it.
            File dbFile = getContext().getDatabasePath(mDatabaseName);
            if (dbFile.delete()) {
                // Perform to create new database.
                onCreate();
            }
        }
    }

    /**
     * <p>
     * </p>
     *
     * @param db
     */
    protected void onDatabaseOpen(SQLiteDatabase db) {
    }

    /**
     * <p>
     * </p>
     */
    protected void onDatabaseCreated() {
    }

    /**
     * Private -------------------------------
     */

    /**
     *
     *
     * @param db
     */
    private void performCreateDatabase(SQLiteDatabase db) {
        // Clear all cached tables.
        if (CACHED_ENTITIES.size() > 0) {
            CACHED_ENTITIES.clear();
        }

        // Dispatch create to all assigned entities but do not cache them yet.
        for (int i = 0; i < ENTITIES.size(); i++) {
            Class<? extends BaseDatabaseEntity> classOfEntity = ENTITIES.get(ENTITIES.keyAt(i));
            try {
                BaseDatabaseEntity table = classOfEntity.newInstance();
                if (table != null) {
                    table.dispatchCreateEntity(db);

                    // Check if entity requests to create some content values.
                    if (table.hasContentValues()) {
                        table.dispatchCreateContentValues(db, getContext());
                    }
                }
            } catch (InstantiationException e) {
                e.printStackTrace();
            } catch (IllegalAccessException e) {
                e.printStackTrace();
            }
        }
        onCreateDatabase(db);
    }

    /**
     *
     *
     * @param db
     * @param oldVersion
     * @param newVersion
     */
    private void performUpgradeDatabase(SQLiteDatabase db, int oldVersion, int newVersion) {
        // TODO: update?
        onUpgradeDatabase(db, oldVersion, newVersion);
    }

	/**
	 *
	 */
	private void dispatchPrepareTables() {
		/**
		 * Assign all annotated entities.
		 */
		final Class<?> providerClass = getClass();
		if (providerClass.isAnnotationPresent(DatabaseEntities.class)) {
			final Class<? extends BaseDatabaseEntity>[] entityClasses = getClass().getAnnotation(DatabaseEntities.class).value();
			for (Class<? extends BaseDatabaseEntity> classOfEntity : entityClasses) {
				assignEntity(classOfEntity);
			}
		}

		// Dispatch to prepare remaining (dynamic) entities.
		onPrepareEntities();
	}

    /**
     *
     *
     * @param db
     */
    private void dispatchDatabaseOpen(SQLiteDatabase db) {
        onDatabaseOpen(db);
    }

    /**
     *
     */
    private void dispatchDatabaseCreated() {
        onDatabaseCreated();
    }

    /**
     *
     * @param entityUri
     * @return
     */
    private BaseDatabaseEntity getEntityByUri(Uri entityUri) {
        BaseDatabaseEntity entity = null;

        final int entityCode = URI_MATCHER.match(entityUri);
        if (entityCode != -1) {
            // Check for cached entity.
            entity = CACHED_ENTITIES.get(entityCode);
            if (entity == null) {
                // Initialize this database entity.
                Class<? extends BaseDatabaseEntity> classOfEntity = ENTITIES.get(entityCode);

                try {
                    // Class should not be NULL when entityCode is correct.
                    entity = classOfEntity.newInstance();
                } catch (InstantiationException e) {
                    e.printStackTrace();
                } catch (IllegalAccessException e) {
                    e.printStackTrace();
                }
            }
        } else {
            Log.e(TAG, "No database entity found with the requested Uri(" + entityUri + ")");
        }
        return entity;
    }

    /**
     *
     * @param entity
     */
    private void cacheEntity(BaseDatabaseEntity entity) {
        // Cache only assigned entity.
        final int entityCode = URI_MATCHER.match(entity.getContentUri());
        if (entityCode != -1) {
            // Check already cached entity.
            if (CACHED_ENTITIES.get(entityCode) == null) {
                // Check if we should clear cached entities.
                if (CACHED_ENTITIES.size() >= sMaxCachedEntitiesCount) {
                    CACHED_ENTITIES.clear();
                }

                // Cache the entity.
                CACHED_ENTITIES.append(entityCode, entity);
            }
        }
    }

    /**
     * Abstract methods ----------------------
     */

    /**
     * Inner classes =========================
     */

    /**
     * <h4>Class Overview</h4>
     * <p>
     * </p>
     *
     * @author Martin Albedinsky
     */
    public static abstract class BaseDatabaseEntity {

        /**
         * Constants =============================
         */

        /**
         * Log TAG.
         */
        private static final String TAG = BaseDatabaseEntity.class.getSimpleName();

        /**
         * Indicates if debug private output trough log-cat is enabled.
         */
        // private static final boolean DEBUG = true;

        /**
         * Indicates if logging for user output trough log-cat is enabled.
         */
        // private static final boolean USER_LOG = true;

        /**
         * Members ===============================
         */

        /**
         * Booleans ------------------------------
         */

        /**
         * Methods ===============================
         */

        /**
         * Public --------------------------------
         */

        /**
         * Protected -----------------------------
         */

        /**
         *
         * @param db
         */
        void dispatchCreateEntity(SQLiteDatabase db) {
            onCreateEntity(db);
        }

        /**
         *
         * @param db
         * @param context
         */
        void dispatchCreateContentValues(SQLiteDatabase db, Context context) {
	        onCreateContentValues(db, context);
        }

        /**
         * Abstract methods ----------------------
         */

        /**
         * <p>
         * </p>
         *
         * @return
         */
        public abstract Uri getContentUri();

        /**
         * <p>
         * </p>
         *
         * @return
         */
        public abstract String getName();

        /**
         * <p>
         * </p>
         *
         * @param db
         * @param projection
         * @param selection
         * @param selectionArgs
         * @param sortOrder
         * @return
         */
        public abstract Cursor dispatchQuery(SQLiteDatabase db, String[] projection, String selection, String[] selectionArgs, String sortOrder);

        /**
         * <p>
         * </p>
         *
         * @param db
         * @param values
         * @return
         */
        public abstract Uri dispatchInsert(SQLiteDatabase db, ContentValues values);

        /**
         * <p>
         * </p>
         *
         * @param db
         * @param values
         * @return
         */
        public abstract int dispatchBulkInsert(SQLiteDatabase db, ContentValues[] values);

        /**
         * <p>
         * </p>
         *
         * @param db
         * @param selection
         * @param selectionArgs
         * @return
         */
        public abstract int dispatchDelete(SQLiteDatabase db, String selection, String[] selectionArgs);

        /**
         * <p>
         * </p>
         *
         * @param db
         * @param values
         * @param selection
         * @param selectionArgs
         * @return
         */
        public abstract int dispatchUpdate(SQLiteDatabase db, ContentValues values, String selection, String[] selectionArgs);

        /**
         * <p>
         * </p>
         *
         * @return
         */
        protected abstract boolean hasContentValues();

        /**
         * <p>
         * </p>
         *
         * @param db
         * @param context
         */
        protected abstract void onCreateContentValues(SQLiteDatabase db, Context context);

        /**
         * <p>
         * </p>
         *
         * @param db
         */
        protected abstract void onCreateEntity(SQLiteDatabase db);
    }

    /**
     * Interface =============================
     */
}
