/*
 * =================================================================================
 * Copyright (C) 2013 Martin Albedinsky [Wolf-ITechnologies]
 * =================================================================================
 * Licensed under the Apache License, Version 2.0 or later (further "License" only);
 * ---------------------------------------------------------------------------------
 * You may use this file only in compliance with the License. More details and copy
 * of this License you may obtain at
 * 
 * 		http://www.apache.org/licenses/LICENSE-2.0
 * 
 * You can redistribute, modify or publish any part of the code written in this
 * file but as it is described in the License, the software distributed under the 
 * License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES or CONDITIONS OF
 * ANY KIND.
 * 
 * See the License for the specific language governing permissions and limitations
 * under the License.
 * =================================================================================
 */
package com.wit.and.database.entity;

import android.content.ContentResolver;
import android.content.ContentValues;
import android.content.Context;
import android.database.Cursor;
import android.database.sqlite.SQLiteDatabase;
import android.database.sqlite.SQLiteQueryBuilder;
import android.net.Uri;
import android.support.v4.content.CursorLoader;
import android.support.v4.content.Loader;
import android.util.Log;
import android.util.SparseArray;

import com.wit.and.database.BaseDatabaseProvider;
import com.wit.and.database.annotations.DatabaseAuthority;
import com.wit.and.database.annotations.entity.EntityColumn;
import com.wit.and.database.annotations.entity.EntityLoader;
import com.wit.and.database.annotations.entity.EntityName;
import com.wit.and.database.annotations.entity.EntityTransactions;
import com.wit.and.database.manage.LoaderOptions;
import com.wit.and.database.manage.TransactionHandler;
import com.wit.and.database.sql.SQLCreateQuery;
import com.wit.and.database.util.DatabaseUtils;

import java.lang.reflect.Field;

/**
 * <h4>Class Overview</h4>
 * <p>
 * Description.
 * </p>
 *
 * @param <TransactionsInterface>
 *
 * @author Martin Albedinsky
 */
public abstract class DatabaseEntity<TransactionsInterface> extends BaseDatabaseProvider.BaseDatabaseEntity {

    /**
     * Constants =============================
     */

    /**
     * Log TAG.
     */
    private static final String TAG = DatabaseEntity.class.getSimpleName();

    /**
     * Indicates if debug private output trough log-cat is enabled.
     */
    // private static final boolean DEBUG = true;

    /**
     * Indicates if logging for user output trough log-cat is enabled.
     */
    // private static final boolean USER_LOG = true;

    /**
     * Enums =================================
     */

    /**
     * Static members ========================
     */

    /**
     * Members ===============================
     */

    /**
     * Name of this database table.
     */
    private String mEntityName;

    /**
     * Uri to content of this database table.
     */
    private Uri mContentUri;

    /**
     *
     */
    private Context mContext;

    /**
     *
     */
    private ContentResolver mContentResolver;

	/**
	 *
	 */
	private TransactionHandler<TransactionsInterface> mTransactioner;

    /**
     * Listeners -----------------------------
     */

    /**
     * Arrays --------------------------------
     */

	/**
	 *
	 */
	private final SparseArray<LoaderOptions> ANNOTATED_LOADERS = new SparseArray<LoaderOptions>();

    /**
     * Booleans ------------------------------
     */

	/**
	 *
	 */
	private boolean bAnnotatedLoadersChecked = false;

    /**
     * Constructors ==========================
     */

    /**
     * <p>
     * </p>
     */
    public DatabaseEntity() {
	    this(true);
    }

    /**
     * <p>
     * </p>
     *
     * @param entityName
     * @param contentUri
     */
    public DatabaseEntity(String entityName, Uri contentUri) {
	    this(false);
        this.mEntityName = entityName;
        this.mContentUri = contentUri;
    }

    /**
     * <p>
     * </p>
     *
     * @param context
     */
    public DatabaseEntity(Context context) {
	    this(null, context);
    }

	/**
	 * <p>
	 * </p>
	 *
	 * @param contentUri
	 * @param context
	 */
	@SuppressWarnings("unchecked")
	public DatabaseEntity(Uri contentUri, Context context) {
		this(true);
		this.mContext = context;
		this.mContentUri = (contentUri != null) ? contentUri : mContentUri;
		this.mContentResolver = (context != null) ? context.getContentResolver() : null;

		// Retrieve table transactions interface.
		final Class<? extends DatabaseEntity> tableClass = getClass();
		if (tableClass.isAnnotationPresent(EntityTransactions.class)) {
			EntityTransactions asyncInterface = tableClass.getAnnotation(EntityTransactions.class);
			this.mTransactioner = new TransactionHandler(this, asyncInterface.value());
		}
	}

	/**
	 *
	 * @param processBaseAnnotations
	 */
	private DatabaseEntity(boolean processBaseAnnotations) {
		final Class<? extends DatabaseEntity> classOfEntity = getClass();
		if (processBaseAnnotations) {
			/**
			 * Process class base annotations.
			 */
			// Retrieve entity name.
			if (classOfEntity.isAnnotationPresent(EntityName.class)) {
				this.mEntityName = classOfEntity.getAnnotation(EntityName.class).value();
			}

			// Retrieve database authority and create entity content URI.
			if (classOfEntity.isAnnotationPresent(DatabaseAuthority.class)) {
				DatabaseAuthority dbAuthority = classOfEntity.getAnnotation(DatabaseAuthority.class);
				this.mContentUri = DatabaseUtils.createDatabaseUri(dbAuthority.value(), mEntityName);
			}
		}
	}

    /**
     * Methods ===============================
     */

    /**
     * Public --------------------------------
     */

	/**
	 * <p>
	 * </p>
	 *
	 * @return
	 */
	public TransactionsInterface transaction() {
		return mTransactioner.prepareTransaction();
	}

	/**
	 * <p>
	 * </p>
	 *
	 * @return
	 */
	public TransactionsInterface transaction(TransactionHandler.TransactionCallback callback) {
		return mTransactioner.prepareTransaction(callback);
	}

    /**
     */
    @Override
    public Cursor dispatchQuery(SQLiteDatabase db, String[] projection, String selection, String[] selectionArgs, String sortOrder) {
        return onQuery(db, projection, selection, selectionArgs, sortOrder);
    }

    /**
     */
    @Override
    public Uri dispatchInsert(SQLiteDatabase db, ContentValues values) {
        Uri rowUri = null;

        db.beginTransaction();
        try {
            rowUri = onInsert(db, values);
            db.setTransactionSuccessful();
        } catch (IllegalStateException e) {
            Log.e(TAG, "Transaction for INSERT action failed.", e);
        } finally {
            db.endTransaction();
        }
        return rowUri;
    }

    /**
     */
    @Override
    public int dispatchBulkInsert(SQLiteDatabase db, ContentValues[] values) {
        int rows = -1;

        db.beginTransaction();
        try {
            rows = onBulkInsert(db, values);
            db.setTransactionSuccessful();
        } catch (IllegalStateException e) {
            Log.e(TAG, "Transaction for BULK INSERT action failed.", e);
        } finally {
            db.endTransaction();
        }
        return rows;
    }

    /**
     */
    @Override
    public int dispatchDelete(SQLiteDatabase db, String selection, String[] selectionArgs) {
        int rows = -1;

        db.beginTransaction();
        try {
            rows = onDelete(db, selection, selectionArgs);
            db.setTransactionSuccessful();
        } catch (IllegalStateException e) {
            Log.e(TAG, "Transaction for DELETE action failed.", e);
        } finally {
            db.endTransaction();
        }
        return rows;
    }

    /**
     */
    @Override
    public int dispatchUpdate(SQLiteDatabase db, ContentValues values, String selection, String[] selectionArgs) {
        int rows = -1;

        db.beginTransaction();
        try {
            rows = onUpdate(db, values, selection, selectionArgs);
            db.setTransactionSuccessful();
        } catch (IllegalStateException e) {
            Log.e(TAG, "Transaction for UPDATE action failed.", e);
        } finally {
            db.endTransaction();
        }
        return rows;
    }

    /**
     * <p>
     * </p>
     *
     * @param projection
     * @param selection
     * @param selectionArgs
     * @param sortOrder
     * @return
     */
    public Cursor query(String[] projection, String selection, String[] selectionArgs, String sortOrder) {
        this.checkContentResolver("QUERY");
        return mContentResolver.query(getContentUri(), projection, selection, selectionArgs, sortOrder);
    }

	/**
	 * <p>
	 * </p>
	 *
	 * @param model
	 * @return
	 */
	public Uri insert(DatabaseEntityModel model) {
		return insert(model.createContentValues());
	}

    /**
     * <p>
     * </p>
     *
     * @param values
     * @return
     */
    public Uri insert(ContentValues values) {
        this.checkContentResolver("INSERT");
        return mContentResolver.insert(getContentUri(), values);
    }

	/**
	 * <p>
	 * </p>
	 *
	 * @param models
	 * @return
	 */
	public int bulkInsert(DatabaseEntityModel[] models) {
		ContentValues[] values = new ContentValues[models.length];
		for (int i = 0; i < values.length; i++) {
			values[i] = models[i].createContentValues();
		}
		return bulkInsert(values);
	}

    /**
     * <p>
     * </p>
     *
     * @param values
     * @return
     */
    public int bulkInsert(ContentValues[] values) {
        this.checkContentResolver("BULK INSERT");
        return mContentResolver.bulkInsert(getContentUri(), values);
    }

    /**
     * <p>
     * </p>
     *
     * @param selection
     * @param selectionArgs
     * @return
     */
    public int delete(String selection, String[] selectionArgs) {
        this.checkContentResolver("DELETE");
        return mContentResolver.delete(getContentUri(), selection, selectionArgs);
    }

	/**
	 * <p>
	 * </p>
	 *
	 * @param model
	 * @return
	 */
	public int update(DatabaseEntityModel model, String selection, String[] selectionArgs) {
		return update(model.createContentValues(), selection, selectionArgs);
	}

    /**
     * <p>
     * </p>
     *
     * @param values
     * @param selection
     * @param selectionArgs
     * @return
     */
    public int update(ContentValues values, String selection, String[] selectionArgs) {
        this.checkContentResolver("UPDATE");
        return mContentResolver.update(getContentUri(), values, selection, selectionArgs);
    }

    /**
     * <p>
     * </p>
     *
     * @param loaderID
     * @return
     */
    public Loader<Cursor> createLoader(int loaderID) {
	    // Check for annotated loader options.
	    this.checkForAnnotatedLoaders();
        return ANNOTATED_LOADERS.indexOfKey(loaderID) >=0 ? createLoader(ANNOTATED_LOADERS.get(loaderID)) : null;
    }

    /**
     * <p>
     * </p>
     *
     * @param options
     * @return
     */
    public Loader<Cursor> createLoader(LoaderOptions options) {
        if (options.isValid()) {
            return new CursorLoader(getContext(), getContentUri(), options.getProjection(), options.getSelection(), options.getSelectionArgs(), options.getSort());
        }
        return null;
    }

    /**
     * Getters + Setters ---------------------
     */

    /**
     */
    @Override
    public String getName() {
        return mEntityName;
    }

    /**
     */
    @Override
    public Uri getContentUri() {
        return mContentUri;
    }

    /**
     * <p>
     * </p>
     *
     * @return
     */
    public ContentResolver getContentResolver() {
        return mContentResolver;
    }

    /**
     * <p>
     * </p>
     *
     * @return
     */
    public Context getContext() {
        return mContext;
    }

    /**
     * Protected -----------------------------
     */

	/**
	 */
	@Override
	protected void onCreateEntity(SQLiteDatabase db) {
		SQLCreateQuery query = new SQLCreateQuery(mEntityName);

		// Retrieve and query all entity columns.
		Field[] fields = getClass().getDeclaredFields();
		for (Field field : fields) {
			if (field.isAnnotationPresent(EntityColumn.class)) {
				EntityColumn entityColumn = field.getAnnotation(EntityColumn.class);

				// Set field accessible to successfully obtain its value.
				field.setAccessible(true);
				try {
					// Build the column.
					DatabaseEntityColumn.ColumnType columnType = entityColumn.value().newInstance();

					// Check for default value.
					if (entityColumn.defaultValue() != null && entityColumn.defaultValue().length() > 0) {
						columnType.setDefaultValue(entityColumn.defaultValue());
					}

					// Check for constraints.
					if (entityColumn.constraints().length > 0) {
						columnType.addConstraints(entityColumn.constraints());
					}
					query.column(new DatabaseEntityColumn(mEntityName + "_" + field.get(this), columnType));
				} catch (IllegalAccessException e) {
					e.printStackTrace();
				} catch (InstantiationException e) {
					e.printStackTrace();
				}
			}
		}

		if (query.hasColumns()) {
			// Create database entity.
			db.execSQL(query.buildQuery());
		}
	}

	/**
     */
    @Override
    protected boolean hasContentValues() {
        return false;
    }

    /**
     */
    @Override
    protected void onCreateContentValues(SQLiteDatabase db, Context context) {
    }

    /**
     * <p>
     * </p>
     *
     * @param db
     * @param projection
     * @param selection
     * @param selectionArgs
     * @param sortOrder
     * @return
     */
    protected Cursor onQuery(SQLiteDatabase db, String[] projection, String selection, String[] selectionArgs, String sortOrder) {
        return this.prepareBuilder().query(db, projection, selection, selectionArgs, null, null, sortOrder);
    }

    /**
     * <p>
     * </p>
     *
     * @param db
     * @param values
     * @return
     */
    protected Uri onInsert(SQLiteDatabase db, ContentValues values) {
        db.insert(getName(), null, values);
        return null;
    }

    /**
     * <p>
     * </p>
     *
     * @param db
     * @param values
     * @return
     */
    protected int onBulkInsert(SQLiteDatabase db, ContentValues[] values) {
        int rows = -1;
        for (ContentValues value : values) {
            db.insert(getName(), null, value);
        }
        return rows;
    }

    /**
     * <p>
     * </p>
     *
     * @param db
     * @param selection
     * @param selectionArgs
     * @return
     */
    protected int onDelete(SQLiteDatabase db, String selection, String[] selectionArgs) {
        return db.delete(getName(), selection, selectionArgs);
    }

    /**
     * <p>
     * </p>
     *
     * @param db
     * @param values
     * @param selection
     * @param selectionArgs
     * @return
     */
    protected int onUpdate(SQLiteDatabase db, ContentValues values, String selection, String[] selectionArgs) {
        return db.update(getName(), values, selection, selectionArgs);
    }

    /**
     * Private -------------------------------
     */

    /**
     *
     * @return
     */
    private SQLiteQueryBuilder prepareBuilder() {
        SQLiteQueryBuilder builder = new SQLiteQueryBuilder();
        builder.setTables(getName());
        return builder;
    }

    /**
     *
     * @param forAction
     */
    private void checkContentResolver(String forAction) {
        if (mContentResolver == null)
            throw new IllegalStateException("Can't perform database action(" + forAction + "). Content resolver is invalid. Don't you forget to create entity with valid Context?");
    }

	/**
	 *
	 */
	private void checkForAnnotatedLoaders() {
		if (bAnnotatedLoadersChecked)
			return;

		/**
		 * Process class loader options.
		 */
		Field[] fields = getClass().getDeclaredFields();
		for (Field field : fields) {
			if (field.isAnnotationPresent(EntityLoader.class)) {
				// Set field accessible to successfully obtain its value.
				field.setAccessible(true);

				// Get loader id.
				final int loaderID = field.getAnnotation(EntityLoader.class).value();

				try {
					// Save loader options.
					ANNOTATED_LOADERS.append(loaderID, (LoaderOptions) field.get(this));
				} catch (IllegalAccessException e) {
					e.printStackTrace();
				}
			}
		}
		// Save check state.
		this.bAnnotatedLoadersChecked = true;
	}

    /**
     * Abstract methods ----------------------
     */

    /**
     * Inner classes =========================
     */

    /**
     * Interface =============================
     */
}
