/*
 * =================================================================================
 * Copyright (C) 2013 Martin Albedinsky [Wolf-ITechnologies]
 * =================================================================================
 * Licensed under the Apache License, Version 2.0 or later (further "License" only);
 * ---------------------------------------------------------------------------------
 * You may use this file only in compliance with the License. More details and copy
 * of this License you may obtain at
 * 
 * 		http://www.apache.org/licenses/LICENSE-2.0
 * 
 * You can redistribute, modify or publish any part of the code written in this
 * file but as it is described in the License, the software distributed under the 
 * License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES or CONDITIONS OF
 * ANY KIND.
 * 
 * See the License for the specific language governing permissions and limitations
 * under the License.
 * =================================================================================
 */
package com.wit.and.database.manage;

import android.os.Handler;
import android.os.Message;
import android.util.Log;

import com.wit.and.database.BaseDatabaseProvider;
import com.wit.and.database.annotations.async.AsyncTransaction;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.LinkedList;
import java.util.Queue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;

/**
 * <h4>Class Overview</h4>
 * <p>
 * </p>
 *
 * @param <TransactionsInterface>
 *
 * @author Martin Albedinsky
 */
public class TransactionHandler<TransactionsInterface> implements InvocationHandler {

	/**
	 * Constants =============================
	 */

	/**
	 * Log TAG.
	 */
	private static final String TAG = com.wit.and.database.manage.TransactionHandler.class.getSimpleName();

	/**
	 * Indicates if debug private output trough log-cat is enabled.
	 */
	// private static final boolean DEBUG = false;

	/**
	 * Indicates if logging for user output trough log-cat is enabled.
	 */
	private static final boolean USER_LOG = true;

	/**
	 * Static members ========================
	 */

	/**
	 * Thread factory for service requests execution.
	 */
	private static final ThreadFactory THREAD_FACTORY = new ThreadFactory() {

		/**
		 */
		@Override
		public Thread newThread(Runnable runnable) {
			return new Thread(runnable, "TransactionResultHandler.Thread");
		}
	};

	/**
	 * Members ===============================
	 */

	/**
	 * Thread executor for service requests execution.
	 */
	private final ExecutorService TRANSACTION_EXECUTOR = Executors.newSingleThreadExecutor(THREAD_FACTORY);

	/**
	 *
	 */
	private TransactionResultHandler mHandler = null;

	/**
	 * Enums =================================
	 */

	/**
	 * Static members ========================
	 */

	/**
	 * Members ===============================
	 */

	/**
	 *
	 */
	private TransactionsInterface mTransactionsInterface;

	/**
	 *
	 */
	private Class<TransactionsInterface> mClassOfTransactionsInterface;

	/**
	 *
	 */
	private BaseDatabaseProvider.BaseDatabaseEntity mDatabaseTable;

	/**
	 * Listeners -----------------------------
	 */

	/**
	 * Arrays --------------------------------
	 */

	/**
	 *
	 */
	private final Queue<TransactionCallback> aRegisteredCallbacks = new LinkedList<TransactionCallback>();

	/**
	 * Booleans ------------------------------
	 */

	/**
	 * Constructors ==========================
	 */

	/**
	 * <p>
	 * </p>
	 *
	 * @param databaseTable
	 * @param classOfTransactionsInterface
	 */
	@SuppressWarnings("unchecked")
	public TransactionHandler(BaseDatabaseProvider.BaseDatabaseEntity databaseTable, Class<TransactionsInterface> classOfTransactionsInterface) {
		this.mDatabaseTable = databaseTable;

		// Set up proxy instance.
		this.mClassOfTransactionsInterface = classOfTransactionsInterface;
		this.mTransactionsInterface = (TransactionsInterface) Proxy.newProxyInstance(databaseTable.getClass().getClassLoader(), new Class<?>[] { classOfTransactionsInterface }, this);
	}

	/**
	 * Methods ===============================
	 */

	/**
	 * Public --------------------------------
	 */

	/**
	 * <p>
	 * <b>Note</b>, that handler will always return <code>null</code> here for the
	 * invoked method with annotation {@link com.wit.and.database.annotations.async.AsyncTransaction}
	 * and the result from the executed transaction will be dispatched to the
	 * {@link com.wit.and.database.manage.TransactionHandler.TransactionCallback},
	 * if was registered.
	 * </p>
	 */
	@Override
	public Object invoke(Object proxy, final Method method, final Object[] args) throws Throwable {
		// Check if the interface can handle this method.
		try {
			mClassOfTransactionsInterface.getDeclaredMethod(method.getName(), method.getParameterTypes());
		} catch (NoSuchMethodException e) {
			throw new NoSuchMethodException(
					"Invoked method(" + method.getName() + ") can't be handled by transaction interface(" + mClassOfTransactionsInterface.getSimpleName() + ") of table(" + mDatabaseTable.getClass().getSimpleName() + ").");
		}

		if (method.isAnnotationPresent(AsyncTransaction.class)) {
			// Create handler if needed.
			if (mHandler == null) {
				// Handler can be created only from the UI thread, so currently invoked method
				// should be invoked from the UI thread, else the exception will be thrown.
				this.mHandler = new TransactionResultHandler();
			}

			// Execute transaction in the background.
			TRANSACTION_EXECUTOR.execute(new Runnable() {
				@Override
				public void run() {
					Object result = null;
					// Try to invoke method on service manager.
					try {
						// Execute service request by invoking service manager method.
						result = method.invoke(mDatabaseTable, args);
					} catch (Exception e) {
						Log.e(TAG, "Error occurred while invoking table transaction interface method(" + method.getName() + ").", e);
					} finally {
						if (USER_LOG) {
							Log.i(TAG, "Database async transaction successfully executed.");
						}
						// Dispatch transaction result on the UI thread.
						mHandler.obtainMessage(TransactionResultHandler.MSG_POST_RESULT, new Result(TransactionHandler.this, result)).sendToTarget();
					}
				}
			});
			return null;
		}

		Object result = method.invoke(proxy, args);
		if (USER_LOG) {
			Log.i(TAG, "Database transaction successfully executed.");
		}
		// Dispatch transaction result on the current thread.
		dispatchTransactionResult(result);
		return result;
	}

	/**
	 * Getters + Setters ---------------------
	 */

	/**
	 * <p>
	 * </p>
	 *
	 * @return
	 */
	public TransactionsInterface prepareTransaction() {
		return mTransactionsInterface;
	}

	/**
	 * <p>
	 * </p>
	 *
	 * @return
	 */
	public TransactionsInterface prepareTransaction(TransactionCallback callback) {
		aRegisteredCallbacks.add(callback);
		return mTransactionsInterface;
	}

	/**
	 * Protected -----------------------------
	 */

	/**
	 * Private -------------------------------
	 */

	/**
	 *
	 * @param result
	 */
	private void dispatchTransactionResult(Object result) {
		TransactionCallback callback = aRegisteredCallbacks.poll();
		if (callback != null) {
			callback.onTransactionExecuted(new TransactionResult(result));
		}
	}

	/**
	 * Abstract methods ----------------------
	 */

	/**
	 * Inner classes =========================
	 */

	/**
	 * <h4>Class Overview</h4>
	 * <p>
	 * </p>
	 *
	 * @author Martin Albedinsky
	 */
	public static class TransactionResult {
		/**
		 * Members ===============================
		 */

		/**
		 *
		 */
		private Object mResult;

		/**
		 * Constructors ==========================
		 */

		/**
		 *
		 * @param result
		 */
		TransactionResult(Object result) {
			this.mResult = result;
		}

		/**
		 * Methods ===============================
		 */

		/**
		 * Public --------------------------------
		 */

		/**
		 * Getters + Setters ---------------------
		 */

		/**
		 * <p>
		 * </p>
		 *
		 * @return
		 */
		public Object getResult() {
			return mResult;
		}

	}

	/**
	 *
	 */
	private static final class TransactionResultHandler extends Handler {
		/**
		 * Constants =============================
		 */

		private static final int MSG_POST_RESULT = 0x01;

		/**
		 * Methods ===============================
		 */

		/**
		 * Public --------------------------------
		 */

		/**
		 */
		@Override
		public void handleMessage(Message msg) {
			switch (msg.what) {
				case MSG_POST_RESULT:
					Result result = (Result) msg.obj;
					if (result != null) {
						result.transactionHandler.dispatchTransactionResult(result.result);
					}
					break;
			}
		}
	}

	/**
	 *
	 */
	private static final class Result {

		/**
		 * Members ===============================
		 */

		/**
		 *
		 */
		private com.wit.and.database.manage.TransactionHandler transactionHandler;

		/**
		 *
		 */
		private Object result;

		/**
		 * Constructors ==========================
		 */

		/**
		 *
		 * @param handler
		 * @param result
		 */
		private Result(com.wit.and.database.manage.TransactionHandler handler, Object result) {
			this.transactionHandler = handler;
			this.result = result;
		}
	}

	/**
	 * Interface =============================
	 */

	/**
	 * <h4>Interface Overview</h4>
	 * <p>
	 * </p>
	 *
	 * @author Martin Albedinsky
	 */
	public static interface TransactionCallback {

		/**
		 * <p>
		 * </p>
		 *
		 * @param result
		 */
		public void onTransactionExecuted(TransactionResult result);
	}
}
